(*
 Sunae Seo, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(* Lex file for C parser(Ansi C) *)
(* Last modification date : 2001/8/10 *)
(* First version is initialized in 2001/8/10, by saseo *)

(* This code is based on ckit developed by Lucent Technologies. *)

{
open Parser
exception Error of string 
exception Fatal_error

fun create_hashtable size init =
  let val tbl = Hashtbl.create size
  in
      List.iter (fn (key, data) => Hashtbl.add tbl key data) init;
      tbl
  end

fun fatal_error msg =
  prerr_string ">> Fatal error: "; prerr_endline msg; raise Fatal_error

fun warning msg =
  prerr_string ">> Warning: "; prerr_endline msg

fun verbose s = if !CpConfig.on_print_lex 
                then (print_string s; print_newline ())

val keyword_table =
  create_hashtable 149 [
    ("break", BREAK),
    ("case", CASE),
    ("continue", CONTINUE),
    ("default", DEFAULT),
    ("do", DO),
    ("else", ELSE),
    ("auto", AUTO),
    ("extern", EXTERN),
    ("register", REGISTER),
    ("static", STATIC),
    ("for", FOR),
    ("goto", GOTO),
    ("if", IF),
    ("char", CHAR),
    ("double", DOUBLE),
    ("enum", ENUM),
    ("float", FLOAT),
    ("int", INT),
    ("long", LONG),
    ("short", SHORT),
    ("struct", STRUCT),
    ("union", UNION),
    ("unsigned", UNSIGNED),
    ("signed", SIGNED),
    ("const", CONST),
    ("volatile", VOLATILE),
    ("void", VOID),
    ("sizeof", SIZEOF),
    ("typedef", TYPEDEF),
    ("return", RETURN),
    ("switch", SWITCH),
    ("while", WHILE)
  ]

fun char_for_backslash c=
  (case c of
      'a' => 7   (* BEL - bel *)
    | 'b' => 8   (* BS - backspace  *)
    | 'f' => 12  (* FF - NP form feed, new page *) 
    | 'n' => 10  (* LF - NL line feed, new line *)
    | 'r' => 13  (* CR - carriage return *)
    | 't' => 9   (* HT - horizontal tab *)
    | 'v' => 11  (* VT - vertical tab *)
    | c => Char.code c )

fun char_for_octal_code lexbuf i maxsize =
  let val c = ref 0
  in
    (for j=0; j<maxsize; j+1
    do
      c := (!c)*8 + (Char.code(Lexing.lexeme_char lexbuf (i+j))-48)
    end);
    if (!c) > 0xFF  then warning "overflow in octal escape sequence";
    (land !c 0xFF ) // Char.chr int_val_of_char
  end

fun char_for_hex_code lexbuf i maxsize = 
  let val c = ref 0
  in
    (for j=0; j<maxsize; j+1
    do
      c := (!c)*16 + (Char.code(Lexing.lexeme_char lexbuf (i+j))-48)
    end);
    if (!c) > 0xFF  then warning "overflow in hex escape sequence";
    (land !c 0xFF )
  end

fun int_of_string_with_l int_val =
  let val pos = (Str.search_forward (Str.regexp "[u U l L]") int_val 0)
                 handle Not_found => -1
  in
    if pos > -1 then int_of_string (String.sub int_val 0 pos) 
    else int_of_string int_val
  end
 
val string_start_pos = ref 0
and comment_start_pos = ref 0

(* To buffer string literals *)

val initial_string_buffer = String.create 256
val string_buff = ref initial_string_buffer
val string_index = ref 0

fun reset_string_buffer () =
  (string_buff := initial_string_buffer;
   string_index := 0)

fun store_string_char c =
  (if !string_index >= String.length (!string_buff) then
   (
    let val new_buff = String.create (String.length (!string_buff) * 2)
    in
      String.blit (!string_buff) 0 new_buff 0 (String.length (!string_buff));
      string_buff := new_buff
    end
   );
   String.unsafe_set (!string_buff) (!string_index) c;
   incr string_index)

fun get_stored_string () =
  let val s = String.sub (!string_buff) 0 (!string_index)
  in
    string_buff := initial_string_buffer; s
  end
}

val id = ['_' 'A'-'Z' 'a'-'z']['_' 'A'-'Z' 'a'-'z' '0'-'9']*
val octdigit = ['0'-'7']
val hexdigit = ['0'-'9' 'a'-'f' 'A'-'F']
val hexnum = '0'['x' 'X'](hexdigit)+['u' 'U' 'l' 'L']?['u' 'U' 'l' 'L']?
val octnum = '0'(octdigit)+['u' 'U' 'l' 'L']?['u' 'U' 'l' 'L']?
val decnum = ('0'|(['1'-'9']['0'-'9']*))['u' 'U' 'l' 'L']?['u' 'U' 'l' 'L']?
val realnum = ((['0'-'9']+('.'['0'-'9']+)?)|('.'['0'-'9']+))(['e' 'E']['+' '-']?['0'-'9']+)?['l' 'L']?
val ws = ('\012'|['\t' ' '])*

val simplecharconst = '\'' [^'\n' '\\'] '\''
val escapecharconst  = '\'' '\\' [^'\n'] '\''

val directive = '#' [^'\n']* '\n'


rule token = parse 
    ws         { token lexbuf } 
  | '\n'       { PrintErr.incr_ln (); token lexbuf}
  | directive  { PrintErr.incr_ln (); token lexbuf } 
  | ":"        { verbose ":"; COLON }      
  | ";"        { verbose ";"; SEMICOLON }      
  | "("        { verbose "("; LPAREN }      
  | ")"        { verbose ")"; RPAREN }      
  | "["        { verbose "["; LBRACE }      
  | "]"        { verbose "]"; RBRACE }      
  | "{"        { verbose "{"; LCURLY }      
  | "}"        { verbose "}"; RCURLY }      
  | "."        { verbose "."; DOT }      
  | "..."      { verbose "..."; ELIPSIS }      
  | ","        { verbose ","; COMMA }      
  | "*"        { verbose "*"; TIMES }      
  | "!"        { verbose "!"; BANG }      
  | "^"        { verbose "^"; HAT }      
  | "+"        { verbose "+"; PLUS }      
  | "-"        { verbose "-"; MINUS }      
  | "++"       { verbose "++"; INC }      
  | "--"       { verbose "--"; DEC }      
  | "->"       { verbose "->"; ARROW }      
  | "/"        { verbose "/"; DIVIDE }      
  | "~"        { verbose "~"; TILDE }      
  | "?"        { verbose "?"; QUESTION }      
  | "|"        { verbose "|"; BAR }      
  | "&"        { verbose "&"; AMP }      
  | "%"        { verbose "%"; PERCENT }      
  | "<="       { verbose "<="; LTE }      
  | ">="       { verbose ">="; GTE }      
  | "=="       { verbose "=="; EQ }      
  | "="        { verbose "="; EQUALS }      
  | "+="       { verbose "+="; PLUSEQUALS }      
  | "-="       { verbose "-="; MINUSEQUALS }      
  | "^="       { verbose "^="; XOREQUALS }      
  | "%="       { verbose "%="; MODEQUALS }      
  | "*="       { verbose "*="; TIMESEQUALS }      
  | "/="       { verbose "/="; DIVEQUALS }      
  | "|="       { verbose "|="; OREQUALS }      
  | "&="       { verbose "&="; ANDEQUALS }      
  | "<<="      { verbose "<<="; LSHIFTEQUALS }      
  | ">>="      { verbose ">>="; RSHIFTEQUALS }      
  | "<"        { verbose "<"; LT }      
  | ">"        { verbose ">"; GT }      
  | "!="       { verbose "!="; NEQ }      
  | "||"       { verbose "||"; OR }      
  | "&&"       { verbose "&&"; AND }      
  | "<<"       { verbose "<<"; LSHIFT }      
  | ">>"       { verbose ">>"; RSHIFT }      

  (* Numeric Literals *)
  | octnum     { verbose "Int_literal";
                 DECNUM (int_of_string_with_l ("0o"^(Lexing.lexeme lexbuf)))}
  | hexnum     { verbose "Int_literal";
                 let val lexval = Lexing.lexeme lexbuf
		     val lexlen = String.length lexval
		 in
                   DECNUM (
		     int_of_string_with_l 
                       ("0x"^(String.sub  lexval 2 (lexlen-2))))
                 end }
  | decnum     { verbose "Int_literal"; 
                 DECNUM (int_of_string (Lexing.lexeme lexbuf))}
  | realnum    { verbose "Real_literal";
                 REALNUM (Lexing.lexeme lexbuf)}

  (* Char *)
  | "'" '\\' ['0'-'3'] ['0'-'7'] ['0'-'7'] "'"
      { verbose "Char_literal";
        CCONST (char_for_octal_code lexbuf 2 3) }
  | "'" '\\' ['0'-'7'] ['0'-'7'] "'"
      { verbose "Char_literal";
        CCONST (char_for_octal_code lexbuf 2 2) }
  | "'" '\\' ['0'-'7'] "'"
      { verbose "Char_literal";
        CCONST (char_for_octal_code lexbuf 2 1) }  
  | "'" '\\' 'x' hexdigit+ "'"
      { verbose "Char_literal";
        let val size = (Lexing.lexeme_end lexbuf)-(Lexing.lexeme_start lexbuf)+1
        in
            CCONST (char_for_hex_code lexbuf 3 (size-4))
        end }
  | simplecharconst 
      { verbose "Char_literal";
        CCONST (Char.code(Lexing.lexeme_char lexbuf 1)) }
  | escapecharconst 
      { verbose "Char_literal";
        let val c = Lexing.lexeme_char lexbuf 2
        in
            CCONST (char_for_backslash c)
        end }

  (* comment *)
  | "/*"        
     { comment_start_pos := Lexing.lexeme_start lexbuf;
        comment lexbuf;
        token lexbuf } 
  | "//" [^ '\n']* '\n'
     { PrintErr.incr_ln (); token lexbuf }

  (* string *)
  | "\""   
     { reset_string_buffer (); 
       let val string_start = Lexing.lexeme_start lexbuf
       in verbose "String_literal";
         string_start_pos := string_start;
         string lexbuf;
         Interop.lexing_field_update_lex_start_pos
          (lexbuf, string_start - (Interop.lexing_field_lex_abs_pos lexbuf));
         STRING (get_stored_string())
       end }

  (* identifiers and keywords *)
  | id 
       {let val s = Lexing.lexeme lexbuf
        in
          verbose s; 
	  (Hashtbl.find keyword_table s) 
	     handle Not_found =>
	       if (TypeDefs.checkTdef s) then (TYPE_NAME s)
	       else  (ID s)
	end  }
  | eof { EOF }
  | _ 
     {fatal_error ("Illegal character")}


and string = parse
    "\""
      { () }
  | '\\' ("\010" | "\013" | "\013\010") [' ' '\009'] *
      { string lexbuf }
  | '\\' ['\\' '\'' '"' 'a' 'b' 'f' 'n' 'r' 't' 'v']
      { store_string_char
         (Char.chr (char_for_backslash(Lexing.lexeme_char lexbuf 1)));
        string lexbuf }
  | '\\' ['0'-'7'] ['0'-'7'] ['0'-'7']
      { store_string_char 
          (Char.chr (char_for_octal_code lexbuf 1 3));
        string lexbuf }
  | '\\' '0'
      { store_string_char (Char.chr 0);
        string lexbuf }
  | '\\' 'x' hexdigit+ 
   { let val size = (Lexing.lexeme_end lexbuf)-(Lexing.lexeme_start lexbuf)+1
     in
      store_string_char 
         (Char.chr (char_for_hex_code lexbuf 2 (size-2)));
      string lexbuf
     end }
  | eof
      { raise (Error "string is not closed with \"") }
  | "\n" 
     { raise (Error "string literal cannot have \\n character") }
  | _
      { store_string_char(Lexing.lexeme_char lexbuf 0);
        string lexbuf }

and comment = parse
    "*/" { () }
  | "\n" { PrintErr.incr_ln (); comment lexbuf }
  | eof  { raise (Error "comment is not terminated!" )}  
  | _    { comment lexbuf }


