/*
 Sunae Seo, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*/

/* Yacc file for c-parser (ansi-C) */
/* Last modification date : 2001/8/10 */
/* First version is initialized in 2001/8/10, by saseo */

/*
 * This grammar is based on ckit developed by Lucent Technologies.
 *
 *  Comments by ckit source code below:
 *
 * Overriding design approach:
 *
 * Accept all legal programs, but possibly some illegal ones at this stage.
 * Do not attempt to make a really tight grammar. Our tools are supposed to
 * work on "correct" C programs (i.e. those that cc -ansi would compile without
 * warnings). Of course, a type checker on the parse tree can report some errors
 * as syntax errors.
 *
 * Note on MARK:
 *
 * externalDecl, statement, and expression are the non-terms that are marked.
 * Compound statements are not separately marked.
 * declarations eventually become either a statement or a externalDecl
 *   if they are outside any function. They are marked accordingly.
 *
 * Note on function definitions:
 *
 * The order of the paramaters will always come from the FuncDecr.
 * The types of the parameter may come from the second declaration list
 * (in K&R style)
 *
 * Note on the structure of the grammar:
 *
 * It is difficult to write a LALR(1) grammar based on the grammar given at
 * the back of the K&R book. The basic difficulty is that both TYPE_NAME and
 * ID are tokens that are strings, but it depends on the context whether
 * a given string is to be treated as an ID or a TYPE_NAME.
 * We have borrowed the solution used in GCC's parser specification. In this
 * scheme, the lexer always return the token TYPE_NAME if a name has been
 * defined as a type name (via a typedef) in an applicable scope. The grammar
 *
 */

%{
open ParseTree 

val unknown_decltype = {storage=[],qualifiers=[],specifiers=[]}:decltype

fun ctypeDecrToTypeName (typ as {qualifiers, specifiers},decr) =
      let fun mkTyp spc = {qualifiers=[], specifiers=[spc]}
   fun addQual q = {qualifiers=q::qualifiers, specifiers=specifiers}
       in case decr
     of VarDecr x => (typ,Some x)
      | PointerDecr x =>
  ctypeDecrToTypeName (mkTyp (Pointer typ),x)
      | ArrayDecr (x,sz) =>
  ctypeDecrToTypeName (mkTyp (Array (sz,typ)),x)
      | FuncDecr (x,lst) =>
  ctypeDecrToTypeName (mkTyp (Function{retType=typ,params=lst}),x)
      | QualDecr (q,decr) =>
  ctypeDecrToTypeName (addQual q, decr)
      | EmptyDecr => (typ, None)
      | EllipsesDecr => (mkTyp Ellipses, Some("**ellipses**"))
      | MARKdeclarator(decr) => ctypeDecrToTypeName(typ, decr)
      end

fun dclr2str dcl = 
  (case ctypeDecrToTypeName ({qualifiers=[],specifiers=[]}, dcl)
     of (_,Some s) => s
      | (_,None) => "")

fun combineDecltypes 
    ( {qualifiers=q1,storage=st1,specifiers=sp1}
      , {qualifiers=q2,storage=st2,specifiers=sp2}) =
  {qualifiers=q1@q2,storage=st1@st2,specifiers=sp1@sp2}  

fun applyPointer (PointerDecr x,rest) = PointerDecr (applyPointer (x,rest))
  | applyPointer (QualDecr (q,x),rest) = QualDecr (q, applyPointer (x,rest))
  | applyPointer (EmptyDecr, rest) = rest
  | applyPointer (_, rest) = rest  
    (* NCH/DBM[6/14/99]: this case can never occur  *)

fun decltypeAddStorage(st, {qualifiers,storage,specifiers}) =
 {qualifiers=qualifiers,storage=st::storage,specifiers=specifiers}

fun decltypeAddQualifiers(qs, {qualifiers,storage,specifiers}) =
 {qualifiers=qs@qualifiers,storage=storage,specifiers=specifiers} 

fun decltypeAddQualifier(q, {qualifiers,storage,specifiers}) =
 {qualifiers=q::qualifiers,storage=storage,specifiers=specifiers}

fun decltypeAddSpecifier(sp, {qualifiers,storage,specifiers}) =
 {qualifiers=qualifiers,storage=storage,specifiers=sp::specifiers}

val decltypeAddAll = combineDecltypes

fun loopQd (q::rst, acc) = loopQd(rst, QualDecr(q, acc))
  | loopQd (nil, acc) = acc

(*********** typedef checking : START *****************************)
(* DBM: major kludge, using TYPEDEF as storage class *)
fun insertDeclNames ({storage,...}: decltype, idl) =
    case storage
      of [TYPEDEF] =>  List.iter (fn (dcl,_) => TypeDefs.addTdef (dclr2str dcl)) idl
       | _ => List.iter (fn (dcl,_) => TypeDefs.addNoTdef (dclr2str dcl)) idl

fun insertFuncName dcl =
  let
    val name = dclr2str dcl
  in
    TypeDefs.addNoTdef name
  end

fun insertFuncParams (FuncDecr (_,params)) = 
     let
       fun getName (ct, dclr) = dclr2str dclr
       val names = List.map getName params
     in
       List.iter TypeDefs.addNoTdef names
     end
  | insertFuncParams (ArrayDecr(dcl,_)) = insertFuncParams dcl
  | insertFuncParams (PointerDecr dcl) = insertFuncParams dcl
  | insertFuncParams _ = ()  (* this is actually an error, but it will be 
                              * caught in BuildAst when processing 
			      *  a PT.FunctionDef *)

(*********** typedef checking : END *****************************)

(* datatype and functions for list-like data*)
type 'a seq = SEQ of 'a list
val emptySeq = SEQ nil
fun singletonSeq x = SEQ[x]
fun addToSeq(x, SEQ yl) = SEQ(x :: yl)  (* add to end of sequence! *)
  (* fun addListToEnd(xl, yl) = SEQ((List.rev xl) @ yl) *)
fun addOptToEnd(None, yl) = yl
  | addOptToEnd(Some x, SEQ yl) = SEQ(x :: yl)
fun seqToList(SEQ yl) = List.rev yl

%}


/* Tokens */
%token  EOF 
%token  COLON  SEMICOLON  LPAREN  RPAREN  LCURLY  RCURLY 
%token  LBRACE  RBRACE  DOT 
%token  COMMA  QUESTION  PERCENT  AMP  BAR  TILDE  DIVIDE  PLUS
%token  MINUS  HAT  BANG  TIMES
%token  INC  DEC  ARROW
%token  <string> ID
%token  EQUALS  PLUSEQUALS  MINUSEQUALS  XOREQUALS  MODEQUALS 
%token  TIMESEQUALS  DIVEQUALS  OREQUALS   ANDEQUALS  LSHIFTEQUALS 
%token  RSHIFTEQUALS
%token  LTE  GTE  LT  GT  EQ  NEQ  OR  AND  LSHIFT  RSHIFT
%token  <int> DECNUM
%token  <string> REALNUM
%token  <string> STRING
%token  <int> CCONST
%token  EXTERN  AUTO  STATIC  REGISTER  CONST  VOLATILE
%token  IF  THEN  ELSE
%token  FOR  DO   SWITCH  CASE  DEFAULT
%token  WHILE  RETURN 
%token  BREAK  CONTINUE  GOTO
%token  CHAR  DOUBLE  ENUM  FLOAT  INT  LONG  SHORT
%token  STRUCT  UNION  UNSIGNED  SIGNED
%token  VOID  SIZEOF  TYPEDEF  UNARY
%token  ELIPSIS
%token  <string> TYPE_NAME

%nonassoc ELSE
%left COMMA
%right EQUALS PLUSEQUALS MINUSEQUALS TIMESEQUALS DIVEQUALS MODEQUALS XOREQUALS OREQUALS ANDEQUALS LSHIFTEQUALS RSHIFTEQUALS
%right QUESTION
%left OR
%left AND
%left BAR
%left HAT
%left AMP
%left EQ NEQ
%left LT GT LTE GTE
%left LSHIFT RSHIFT
%left PLUS MINUS
%left TIMES DIVIDE PERCENT
%right UNARY
%right INC DEC SIZEOF
%left LBRACE LPAREN ARROW DOT

%start goal
%type <ParseTree.externalDecl list> goal

%%

goal:  
 tu   {seqToList $1}

tu:
                          {emptySeq}
 | tu externalDeclaration {addOptToEnd($2, $1)}

externalDeclaration:
   declaration         {Some(MARKexternalDecl (ExternalDecl $1))} 
 | SEMICOLON           {None}
 | functionDefinition  {Some (MARKexternalDecl $1)}

statement:
   FOR LPAREN opExpr SEMICOLON opExpr SEMICOLON opExpr RPAREN statement
                           {MARKstatement (For($3,$5,$7,$9))}
 | WHILE LPAREN exprWComma RPAREN statement
                           {MARKstatement (While($3,$5))}
 | SWITCH LPAREN exprWComma RPAREN statement
                           {MARKstatement (Switch($3,$5))}
 | DO statement WHILE LPAREN exprWComma RPAREN SEMICOLON
                           {MARKstatement (Do($5,$2))}
 | BREAK SEMICOLON         {MARKstatement Break}
 | CONTINUE SEMICOLON      {MARKstatement Continue} 
 | RETURN opExpr SEMICOLON {MARKstatement (Return $2)}
 | GOTO ID SEMICOLON       {MARKstatement (Goto $2)}
 | compoundStatement       {$1}
 | ID COLON statement      {MARKstatement (Labeled ($1,$3))}
 | DEFAULT COLON statement {MARKstatement (DefaultLabel($3))}
 | CASE exprWComma COLON statement  
                           {MARKstatement (CaseLabel($2,$4))}
 | IF LPAREN exprWComma RPAREN statement
                           {MARKstatement (IfThen($3,$5))}
 | IF LPAREN exprWComma RPAREN statement ELSE statement
                           {MARKstatement (IfThenElse($3,$5,$7))}
 | exprWComma SEMICOLON    {MARKstatement (Expr($1))}
 | SEMICOLON               {MARKstatement (Expr(EmptyExpr))}



declaration:
  declaration1 SEMICOLON        {$1}


declaration1:
   declarationSpecifiers
     {insertDeclNames ($1, []); 
      MARKdeclaration (Declaration ($1,[]))}

 | declarationSpecifiers initDeclaratorList
     {let val decl = ($1, seqToList $2)
      in insertDeclNames decl;
         MARKdeclaration (Declaration decl) end}

 | declarationModifiers notypeInitDeclaratorList
     {let val decl = ($1, seqToList $2)
      in insertDeclNames decl;
         MARKdeclaration (Declaration decl) end}

ostatementlist:  
   statementlist  {seqToList $1}
 |                {[]}

statementlist:
   statement   {singletonSeq $1}
 | statementlist statement {addToSeq($2, $1)}


compoundStatement: 
   LCURLY pushScope declarationList ostatementlist popScope RCURLY
     {Compound ((List.map (fn x => Decl x) (seqToList $3)) @ $4)}
 | LCURLY ostatementlist RCURLY {Compound $2}




unaryOperator: 
   AMP   {AddrOf}
 | TIMES {Star}
 | PLUS  {Uplus}
 | MINUS {Negate}
 | TILDE {BitNot}
 | BANG  {Not}

expr:
   expr QUESTION expr COLON expr %prec QUESTION 
     {MARKexpression (QuestionColon($1,$3,$5))}
 | expr PLUSEQUALS expr   {MARKexpression (Binop(PlusAssign,$1,$3))}
 | expr MINUSEQUALS expr  {MARKexpression (Binop(MinusAssign,$1,$3))}
 | expr TIMESEQUALS expr  {MARKexpression (Binop(TimesAssign,$1,$3))}
 | expr DIVEQUALS expr    {MARKexpression (Binop(DivAssign,$1,$3))}
 | expr MODEQUALS expr    {MARKexpression (Binop(ModAssign,$1,$3))}
 | expr XOREQUALS expr    {MARKexpression (Binop(XorAssign,$1,$3))}
 | expr OREQUALS expr     {MARKexpression (Binop(OrAssign,$1,$3))}
 | expr ANDEQUALS expr    {MARKexpression (Binop(AndAssign,$1,$3))}
 | expr LSHIFTEQUALS expr {MARKexpression (Binop(LshiftAssign,$1,$3))}
 | expr RSHIFTEQUALS expr {MARKexpression (Binop(RshiftAssign,$1,$3))}
 | expr EQUALS expr       {MARKexpression (Binop(Assign,$1,$3))}
 | expr OR expr           {MARKexpression (Binop(Or,$1,$3))}
 | expr AND expr          {MARKexpression (Binop(And,$1,$3))}
 | expr BAR expr          {MARKexpression (Binop(BitOr,$1,$3))}
 | expr HAT expr          {MARKexpression (Binop(BitXor,$1,$3))}
 | expr AMP expr          {MARKexpression (Binop(BitAnd,$1,$3))}
 | expr EQ expr           {MARKexpression (Binop(Eq,$1,$3))}
 | expr NEQ expr          {MARKexpression (Binop(Neq,$1,$3))}
 | expr LT expr           {MARKexpression (Binop(Lt,$1,$3))}
 | expr GT expr           {MARKexpression (Binop(Gt,$1,$3))}
 | expr LTE expr          {MARKexpression (Binop(Lte,$1,$3))}
 | expr GTE expr          {MARKexpression (Binop(Gte,$1,$3))}
 | expr LSHIFT expr       {MARKexpression (Binop(Lshift,$1,$3))}
 | expr RSHIFT expr       {MARKexpression (Binop(Rshift,$1,$3))}
 | expr PLUS expr         {MARKexpression (Binop(Plus,$1,$3))}
 | expr MINUS expr        {MARKexpression (Binop(Minus,$1,$3))}
 | expr TIMES expr        {MARKexpression (Binop(Times,$1,$3))}
 | expr DIVIDE expr       {MARKexpression (Binop(Divide,$1,$3))}
 | expr PERCENT expr      {MARKexpression (Binop(Mod,$1,$3))}
 | expr INC %prec INC     {MARKexpression (Unop(PostInc,$1))}
 | expr DEC %prec INC     {MARKexpression (Unop(PostDec,$1))}
 | INC expr %prec INC     {MARKexpression (Unop(PreInc,$2))}
 | DEC expr %prec INC     {MARKexpression (Unop(PreDec,$2))}
 | unaryOperator expr  %prec UNARY {MARKexpression (Unop($1,$2))}
 | SIZEOF expr            {MARKexpression (Unop(Sizeof,$2))}
 | LPAREN typeName RPAREN expr %prec INC {MARKexpression (Cast ($2,$4))}
 | SIZEOF LPAREN typeName RPAREN %prec SIZEOF 
                          {MARKexpression (Unop(SizeofType $3,EmptyExpr))}
 | expr LBRACE exprWComma RBRACE {MARKexpression (Binop(Sub,$1,$3))}


 | expr LPAREN RPAREN  {MARKexpression (Call($1,[]))}
 | expr LPAREN argumentExprList RPAREN
                       {MARKexpression (Call($1, seqToList $3))}
 | expr DOT ID         {MARKexpression (Binop(Dot,$1,Id($3)))}
 | expr ARROW ID       {MARKexpression (Binop(Arrow,$1,Id($3)))}
 | expr DOT TYPE_NAME  {MARKexpression (Binop(Dot,$1,Id($3)))}
 | expr ARROW TYPE_NAME  {MARKexpression (Binop(Arrow,$1,Id($3)))}
 | LPAREN exprWComma RPAREN {MARKexpression ($2)}
 | DECNUM              {MARKexpression (IntConst $1)}
 | REALNUM             {MARKexpression (RealConst $1)}
 | CCONST              {MARKexpression (IntConst $1)}
 | ID                  {MARKexpression (Id($1))}
 | strings             {MARKexpression (String($1))}


strings:  STRING       {$1}
 | STRING strings      {$1 ^ $2}



exprWComma:
   expr                   {$1}
 | exprWComma COMMA expr  {MARKexpression (Binop(Comma,$1,$3))}

opExpr:         {EmptyExpr}
 | exprWComma   {$1}


argumentExprList:
   expr                          {singletonSeq $1}
 | argumentExprList COMMA expr   {addToSeq($3, $1)}

typeName:
   specifierQualifierList   {$1}
 | specifierQualifierList abstractDeclarator
     {let val (ty,decl) = ctypeDecrToTypeName ($1, $2)
      in ty end}

declarationSpecifiers:
   typeSpecifier reservedDeclarationSpecifier
     {decltypeAddSpecifier ($1, $2)}
 | declarationModifiers typeSpecifier reservedDeclarationSpecifier
     {decltypeAddAll ($1, decltypeAddSpecifier ($2,$3))}

reservedDeclarationSpecifier:
                                        {unknown_decltype}
 | reservedDeclarationSpecifier specifierQualifierReserved
     {let val {qualifiers,specifiers} = $2
          val decltype = 
              {qualifiers=qualifiers,specifiers=specifiers,storage=[]}
      in decltypeAddAll (decltype, $1) end}
 | reservedDeclarationSpecifier storageClassSpecifier
     {decltypeAddStorage ($2,$1)}

specifierQualifierReserved:
   typeSpecifierReserved       {{qualifiers=[],specifiers=[$1]}}
 | typeQualifier               {{qualifiers=[$1],specifiers=[]}}
 | structOrUnionSpecifier      {{qualifiers=[],specifiers=[$1]}}
 | enumSpecifier               {{qualifiers=[],specifiers=[$1]}}

declarationModifiers:
   storageClassSpecifier
     {{storage = [$1], qualifiers = [], specifiers = []}}
 | declarationModifiers storageClassSpecifier
     {decltypeAddStorage($2,$1)}
 | typeQualifier
     {{specifiers = [], storage = [], qualifiers = [$1]}}
 | declarationModifiers typeQualifier
     {decltypeAddQualifier($2, $1)}

specifierQualifierList:
   typeSpecifier reservedSpecifierQualifiers
     {let val {specifiers, qualifiers} = $2
      in {specifiers=$1::specifiers,qualifiers=qualifiers} end}
             
 | typeQualifierList typeSpecifier reservedSpecifierQualifiers
     {let val {specifiers, qualifiers} = $3
      in {specifiers=$2::specifiers,qualifiers=$1@qualifiers} end}

reservedSpecifierQualifiers:
     {{qualifiers=[],specifiers=[]}}
 | reservedSpecifierQualifiers specifierQualifierReserved
     {let val {specifiers=s1, qualifiers=q1} = $1
          val {specifiers=s2, qualifiers=q2} = $2
      in {specifiers=s1@s2, qualifiers=q1@q2} end}
  
typeQualifierList:
   typeQualifier                     {[$1]}
 | typeQualifier typeQualifierList   {$1::$2}

typeSpecifier:
   typeSpecifierReserved       {$1}
 | structOrUnionSpecifier      {$1}
 | enumSpecifier               {$1}
 | TYPE_NAME                   {TypedefName $1}

typeSpecifierReserved:
   VOID      {Void}
 | CHAR      {Char}
 | SHORT     {Short}
 | INT       {Int}
 | LONG      {Long}
 | FLOAT     {Float}
 | DOUBLE    {Double}
 | SIGNED    {Signed}
 | UNSIGNED  {Unsigned}


structOrUnionSpecifier:
   structOrUnion LCURLY structDeclarationList RCURLY
     {Struct{isStruct=$1, tagOpt=None, members=seqToList $3}}
 | structOrUnion ID LCURLY structDeclarationList RCURLY
     {Struct{isStruct=$1, tagOpt=Some $2, members=seqToList $4}}
 | structOrUnion TYPE_NAME LCURLY structDeclarationList RCURLY
     {Struct{isStruct=$1, tagOpt=Some $2, members=seqToList $4}}
 | structOrUnion ID   {StructTag {isStruct=$1, name=$2}}
 | structOrUnion TYPE_NAME  {StructTag {isStruct=$1, name=$2}}

   /* humor me: consider true for struct */
structOrUnion:
   STRUCT    {true}
 | UNION    {false}

structDeclarationList:
   structDeclaration                       {singletonSeq $1}
 | structDeclarationList structDeclaration {addToSeq($2, $1)}

structDeclaration:
   specifierQualifierList structDeclaratorList SEMICOLON
     {($1, seqToList $2)}
 | typeQualifierList notypeStructDeclaratorList SEMICOLON
     {let val ct = {qualifiers=$1, specifiers=[]}
      in (ct, seqToList $2) end}

structDeclaratorList:
   structDeclarator                            {singletonSeq $1}
 | structDeclaratorList COMMA structDeclarator {addToSeq($3, $1)}

notypeStructDeclaratorList:
   notypeStructDeclarator                            {singletonSeq $1}
 | notypeStructDeclaratorList COMMA structDeclarator {addToSeq($3, $1)}

structDeclarator:
   declarator             {($1, EmptyExpr)}
 | COLON expr             {(EmptyDecr, $2)}
 | declarator COLON expr  {($1, $3)}

notypeStructDeclarator:
   notypeDeclarator            {($1, EmptyExpr)}
 | COLON expr                  {(EmptyDecr, $2)}
 | notypeDeclarator COLON expr {($1, $3)}

typeQualifier:
   CONST      {CONST}
 | VOLATILE   {VOLATILE}

enumSpecifier:
   ENUM LCURLY enumeratorList trailingComma RCURLY
     {Enum{tagOpt=None, enumerators=seqToList $3, trailingComma=$4}}
 | ENUM ID LCURLY enumeratorList trailingComma RCURLY 
     {Enum{tagOpt=Some($2), enumerators=seqToList $4, trailingComma=$5}}
 | ENUM TYPE_NAME LCURLY enumeratorList trailingComma RCURLY 
     {Enum{tagOpt=Some($2), enumerators=seqToList $4, trailingComma=$5}}
 | ENUM ID   {EnumTag($2)}
 | ENUM TYPE_NAME  {EnumTag($2)}

enumeratorList:
   enumeratorList COMMA enumerator 
     {addToSeq($3, $1)}
 | enumerator   
     {let val (id,e) = $1 
      in TypeDefs.addNoTdef(id); singletonSeq $1
      end}

enumerator:
   ID              {($1,EmptyExpr)}
 | ID EQUALS expr  {($1,$3)}

storageClassSpecifier: 
   EXTERN    {EXTERN}
 | STATIC    {STATIC}
 | AUTO      {AUTO}
 | REGISTER  {REGISTER}
 | TYPEDEF   {TYPEDEF}

trailingComma:    {false}
 | COMMA    {true}
 
initDeclaratorList:
   initDeclarator                          {singletonSeq $1}
 | initDeclaratorList COMMA initDeclarator {addToSeq($3, $1)}

initDeclarator:
   declarator                    {($1,EmptyExpr)}
 | declarator EQUALS initializer {($1,$3)}

notypeInitDeclaratorList:
   notypeInitDeclarator                          {singletonSeq $1}
 | notypeInitDeclaratorList COMMA initDeclarator {addToSeq($3, $1)}

notypeInitDeclarator:
   notypeDeclarator                      {($1,EmptyExpr)}
 | notypeDeclarator EQUALS initializer   {($1,$3)}

declarator:
   aftertypeDeclarator          {$1}
 | notypeDeclarator             {$1}

aftertypeDeclarator:
   aftertypeDirectDeclarator         {$1}
 | pointer aftertypeDirectDeclarator {applyPointer($1,$2)}

notypeDeclarator:
   notypeDirectDeclarator         {$1}
 | pointer notypeDirectDeclarator {applyPointer($1, $2)}

parmDeclarator:
   parmDirectDeclarator         {$1}
 | pointer parmDirectDeclarator {applyPointer($1, $2)}

pointer:
   TIMES                           {PointerDecr(EmptyDecr)}
 | TIMES typeQualifierList         {PointerDecr(loopQd($2,EmptyDecr))}
 | TIMES pointer                   {PointerDecr($2)}
 | TIMES typeQualifierList pointer {PointerDecr(loopQd($2,$3))}

aftertypeDirectDeclarator:
   TYPE_NAME                             {VarDecr $1}
 | LPAREN aftertypeDeclarator RPAREN     {$2}
 | aftertypeDirectDeclarator  LBRACE RBRACE %prec DOT 
     {ArrayDecr ($1,EmptyExpr)}
 | aftertypeDirectDeclarator LBRACE expr RBRACE %prec DOT 
     {ArrayDecr ($1,$3)}
 | aftertypeDirectDeclarator LPAREN RPAREN  %prec DOT 
     {FuncDecr ($1,nil)}
 | aftertypeDirectDeclarator LPAREN parameterTypeList RPAREN   %prec DOT
     {FuncDecr ($1,$3)}
 | aftertypeDirectDeclarator LPAREN identlist RPAREN  %prec DOT
     {FuncDecr ($1, 
         List.map (fn x => (unknown_decltype,VarDecr x)) (seqToList $3))}

notypeDirectDeclarator:
   ID    {VarDecr $1}
 | LPAREN notypeDeclarator RPAREN  {$2}
 | notypeDirectDeclarator LBRACE RBRACE %prec DOT
     {ArrayDecr ($1,EmptyExpr)}
 | notypeDirectDeclarator LBRACE expr RBRACE  %prec DOT
     {ArrayDecr ($1,$3)}
 | notypeDirectDeclarator LPAREN RPAREN  %prec DOT 
     {FuncDecr ($1,nil)}
 | notypeDirectDeclarator LPAREN parameterTypeList RPAREN  %prec DOT 
     {FuncDecr ($1,$3)}
 | notypeDirectDeclarator LPAREN identlist RPAREN  %prec DOT 
     {FuncDecr ($1,
         List.map (fn x => (unknown_decltype,VarDecr x)) (seqToList $3))}

parmDirectDeclarator:
   TYPE_NAME   {VarDecr $1}
 | parmDirectDeclarator LBRACE RBRACE  %prec DOT
     {ArrayDecr ($1,EmptyExpr)}
 | parmDirectDeclarator LBRACE expr RBRACE  %prec DOT
     {ArrayDecr ($1,$3)}
 | parmDirectDeclarator LPAREN RPAREN   %prec DOT
     {FuncDecr ($1,nil)}
 | parmDirectDeclarator LPAREN parameterTypeList RPAREN  %prec DOT 
     {FuncDecr ($1,$3)}
 | parmDirectDeclarator LPAREN identlist RPAREN  %prec DOT 
     {FuncDecr ($1,
          List.map (fn x => (unknown_decltype,VarDecr x)) (seqToList $3))}

initializer:
   expr    {$1}
 | LCURLY initializerList trailingComma RCURLY
     {MARKexpression (InitList(seqToList $2))}

initializerList:
   initializer   {singletonSeq $1}
 | initializerList COMMA initializer {addToSeq($3, $1)}

declarationList:
   declaration          
     {singletonSeq(MARKdeclaration $1)}
 | declarationList declaration 
     {addToSeq(MARKdeclaration $2, $1)}
   
identlist:
   ID                 {singletonSeq $1}
 | identlist COMMA ID {addToSeq($3, $1)}

/* Put function name in the current scope and param names in a pushed scope. */
fDefDeclaration:
   notypeDeclarator
     {insertFuncName($1);
      TypeDefs.pushScope();
      insertFuncParams($1);
      (unknown_decltype, $1)}
 | declarationSpecifiers declarator
     {insertFuncName($2);
      TypeDefs.pushScope();
      insertFuncParams($2);
      ($1, $2)}
 | declarationModifiers notypeDeclarator
     {insertFuncName($2);
      TypeDefs.pushScope();
      insertFuncParams($2);
      ($1, $2)}

functionDefinition:
   fDefDeclaration compoundStatement
     {TypeDefs.popScope(); 
      let val (ret,fnname) = $1
      in 
         FunctionDef
            {retType = ret,
             funDecr = fnname,
             krParams = [],
             body = $2} end}
     
 | fDefDeclaration declarationList compoundStatement
     {TypeDefs.popScope();
      let val (ret, fnname) = $1
      in
         FunctionDef 
            {retType = ret,
             funDecr = fnname,
             krParams = seqToList $2,
             body = $3} end}
     
abstractDeclarator:
   pointer                          {applyPointer ($1, EmptyDecr)}
 | directAbstractDeclarator         {$1}
 | pointer directAbstractDeclarator {applyPointer($1, $2)}

directAbstractDeclarator:
   LPAREN abstractDeclarator RPAREN            {$2}
 | LBRACE RBRACE                               {ArrayDecr(EmptyDecr, EmptyExpr)}
 | LBRACE expr RBRACE                          {ArrayDecr(EmptyDecr, $2)}
 | directAbstractDeclarator LBRACE RBRACE      {ArrayDecr ($1,EmptyExpr)}
 | directAbstractDeclarator LBRACE expr RBRACE {ArrayDecr ($1,$3)}
 | LPAREN RPAREN                               {FuncDecr (EmptyDecr ,nil)}
 | LPAREN parameterTypeList RPAREN             {FuncDecr (EmptyDecr, $2)}
 | directAbstractDeclarator LPAREN RPAREN      {FuncDecr ($1,nil)}
 | directAbstractDeclarator LPAREN parameterTypeList RPAREN 
                                               {FuncDecr ($1, $3)}

parameterTypeList:
   parameterList                         {seqToList $1}
 | parameterList COMMA ELIPSIS
    {let val decltype = {specifiers=[Ellipses],qualifiers=[],storage=[]}
     in (seqToList $1) @ [(decltype, EllipsesDecr)] end}

parameterList:
   parameterDeclaration 
     {let val (ty,decl) = $1
      in singletonSeq(ty, MARKdeclarator decl) end}
 | parameterList COMMA parameterDeclaration 
     {let val (ty,decl) = $3
      in addToSeq((ty, MARKdeclarator decl), $1) end}

/* Decided not a push and pop a scope at the parameterDeclarations, because we
 * are not going to directly plug these names in the tdef table. If this is just
 * a function declaration, the names here do not matter to the tdef table. If
 * this will be part of func definition, we put all these names in tdef table at
 * proper scope correctly, later on.
 * Note: We miss syntax errors like int f(int foo, foo bar);, if foo was a typename
 */
parameterDeclaration:
   declarationSpecifiers  notypeDeclarator   {($1, $2)}
 | declarationSpecifiers  parmDeclarator     {($1, $2)}
 | declarationSpecifiers  /* this case can arise for function prototypes */
     {($1, EmptyDecr)}
 | declarationSpecifiers  abstractDeclarator {($1, $2)}
 | declarationModifiers  notypeDeclarator    {($1, $2)}
 | declarationModifiers  abstractDeclarator  {($1, $2)}

pushScope:    {TypeDefs.pushScope()}

popScope:     {TypeDefs.popScope()}


