(*
 Sunae Seo, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

(* This code is based on ckit developed by Lucent Technologies. *)

signature TYPEDEFS =
sig
  val addTdef: string -> unit    
	(* a string as a typename in current scope *)
  val addNoTdef: string -> unit  
    	(* a string is not a typename, 
         * may hide typenames of outer scopes *)
  val checkTdef: string -> bool  
    	(* is this string a typename in current context ? *)
  val reset: unit -> unit        
    	(* clear all tables, needed if you are doing many files *)
  val truncTo: int ref           (* limited-width names ? *)
  val pushScope: unit -> unit    (* entering a new scope in C *)
  val popScope: unit -> unit     (* exiting the last scope *)
end 

(* We need a stack of tables to properly handle the scoping in typenames
 * Remember, there are four type of things competing in the namespace of
 * typenames:
 *   typenames, variables, functions and enum constants
 * Once you enter a new scope, reuse of these names can hide previous
 * uses.
 * Also note that struct field names do not redefine names within their scope
 * So, the following is legal:
 *    typedef int bar;
 *    struct h {
 *      bar bar;
 *      bar baz;
 *    };
 *)

structure TypeDefs : TYPEDEFS = 
struct
  val truncTo: int ref = ref CpConfig.symbol_length
  type item = bool (* true says typename, false says else *)
  val sayTdefs : bool ref = ref true
  val tdefTable: (((string, bool) Hashtbl.t) list) ref 
              = ref [Hashtbl.create 1024]

  fun checkTdef str =
    let 
      val name = (String.sub str 0 (!truncTo)) handle Invalid_argument s => str
      fun lookup  n  nil  = false
        | lookup  n (fst::rst) = 
	     (Hashtbl.find fst n) handle Not_found => (lookup n rst)
    in
      (lookup name (!tdefTable)) andalso (!sayTdefs) 
    end


  fun pushScope () = 
     (tdefTable := (Hashtbl.create 1024)::(!tdefTable))

  fun popScope () =   
      (case (!tdefTable) of 
        [x] => ()  (* don't change *)
        | (_ :: l) => (tdefTable := l)
        | nil => ())
    (* don't change; but we are in trouble here! *)
 
  val errorCount = ref 0

  fun reset() = (tdefTable := [Hashtbl.create 1024]; errorCount := 0)

    (* TBD: In the next two functions, it is an option to raise a syntax error,
     * if there is a redefinition in the same scope, i.e., the topmost table
     *)
      
  fun addTdef str = 
    let
      val name = (String.sub str 0 (!truncTo)) handle Invalid_argument s => str
    in
      (* insert name in the top of tdefTable as a typename *)
      case !tdefTable of
        x :: _ => Hashtbl.add x name true
        | nil => 
	  (if !errorCount = 0
           then prerr_string "Error: empty type def table (lexer), probably caused by syntax error";
           (* should be Error.error, but don't have an error stream handy. *)
          errorCount := !errorCount + 1)
    end

  fun addNoTdef str = 
    let
      val name = (String.sub str 0 (!truncTo)) handle Invalid_argument s => str
    in
      (* insert name in the top of tdefTable as not a typename *)
      case !tdefTable of
        x :: _ => Hashtbl.add x name false
        | nil => 
	  (if !errorCount = 0
           then prerr_string "Error: empty type def table (lexer), probably caused by syntax error";
           (* should be Error.error, but don't have an error stream handy. *)
           errorCount := !errorCount + 1)
    end
    
end

